
#ifdef DERIVATIVES    
    #extension GL_OES_standard_derivatives : enable
#endif

precision mediump float;
uniform sampler2D uTex0; //sdfTexture
uniform vec2      uTexSize;//sdf Texture Size
uniform vec4      uFgColor;

uniform float     uGamma;
uniform float	  uOutlineWidth;
uniform vec4 	  uOutlineColor;
uniform float     uOutlineAlpha;

varying vec2      textureCoordinate;

float median( float r, float g, float b )
{
	return max( min( r, g ), min( max( r, g ), b ) );
}

vec2 safeNormalize( in vec2 v )
{
   float len = length( v );
   len = ( len > 0.0 ) ? 1.0 / len : 0.0;
   return v * len;
}

float calcDiff(vec2 p)
{
    return p.x * p.x - p.y;
}


void main(void) {
    // Sample SDFA texture (4 channels).
    vec4 sample = texture2D( uTex0, textureCoordinate);
    // Calculate signed distance (in texels).
    float sigDist = median( sample.r, sample.g, sample.b ) - 0.5;

#ifdef DERIVATIVES    
    // Convert normalized texcoords to absolute texcoords.
    vec2 uv = textureCoordinate * uTexSize;
    // Calculate derivates
    vec2 Jdx = dFdx( uv );
    vec2 Jdy = dFdy( uv );
    
    // For proper anti-aliasing, we need to calculate signed distance in pixels. We do this using derivatives.
    vec2 gradDist = safeNormalize( vec2( dFdx( sigDist ), dFdy( sigDist ) ) );
    vec2 grad = vec2( gradDist.x * Jdx.x + gradDist.y * Jdy.x, gradDist.x * Jdx.y + gradDist.y * Jdy.y );
    // Apply anti-aliasing.
    const float kThickness = 0.125;
    float kNormalization = kThickness * 0.5 * sqrt( 2.0 );
    float afwidth = min( kNormalization * length( grad ), 0.5 );    
#else
    float c = calcDiff(textureCoordinate);
    vec2 ps = vec2( 1.0 / uTexSize.x, 1.0 / uTexSize.y );
    float dfdx = calcDiff( textureCoordinate + vec2( ps.x ) ) - c;
    float dfdy = calcDiff( textureCoordinate + vec2( ps.y ) ) - c;
    float afwidth = abs( dfdx ) + abs( dfdy );  
#endif    

    float opacity = smoothstep( 0.0 - afwidth, 0.0 + afwidth, sigDist );
    vec4 color = uFgColor;
    color.a = uFgColor.a * pow(opacity, 1.0 / uGamma);
    
    if(uOutlineWidth > 0.01){
    	float outlineWidth = uOutlineWidth * 0.445;
    	outlineWidth = min(outlineWidth, 0.49 - afwidth);
    	float border = pow(opacity, 1.0 / uGamma);
    	opacity = smoothstep(-outlineWidth-afwidth, outlineWidth + afwidth, sample.a - 0.5);
    	color.rgb = mix(uOutlineColor.rgb, uFgColor.rgb, border);
        color.a = uOutlineAlpha * uFgColor.a * pow(opacity, 1.0 / uGamma);
    }

    
    gl_FragColor = color;
}