local init_state = 1
local text_state = 0

exports = {}
local ani_load_status = 0
local cur_animation_path = nil
local last_animation_path = nil

local style_load_status = 0
local cur_style_path = nil
local last_style_path = nil

local ge_feature_name = "FontRender/"
local shadow_pass_name = "textShadow"
local text_pass_name = "text"
local text_feature_name = "IS_Text/"

local reforce_set_render_options = false

local render_param_json = nil 


local function printlog(log_info)
    globals.lua_log(string.format("Lua message: %s", log_info));
end

function printTable ( t )
    local print_r_cache={}
    local function sub_print_r(t,indent)
        if (print_r_cache[tostring(t)]) then
            printlog(indent.."*"..tostring(t))
        else
            print_r_cache[tostring(t)]=true
            if (type(t)=="table") then
                for pos,val in pairs(t) do
                    if (type(val)=="table") then
                        printlog(indent.."["..pos.."] => "..tostring(t).." {")
                        sub_print_r(val,indent..string.rep(" ",string.len(pos)+8))
                        printlog(indent..string.rep(" ",string.len(pos)+6).."}")
                    elseif (type(val)=="string") then
                        printlog(indent.."["..pos..'] => "'..val..'"')
                    else
                        printlog(indent.."["..pos.."] => "..tostring(val))
                    end
                end
            else
                printlog(indent..tostring(t))
            end
        end
    end
    if (type(t)=="table") then
        printlog(tostring(t).." {")
        sub_print_r(t,"  ")
        printlog("}")
    else
        sub_print_r(t,"  ")
    end
    printlog()
end

local init_once = true

function setDuration(dur)
	if math.abs(dur - duration) > 0.01 then
		duration = dur
	end
end

local function clamp(v, minValue, maxValue)  
    if v < minValue then
        return minValue
    end
    if( v > maxValue) then
        return maxValue
    end
    return v 
end

local function setSDFTextureSize(textFeature, renderFeature)
	if not renderFeature or not renderFeature then
		return
	end
	
	local geFeature = EffectSdk.castGeneralEffectFeature(renderFeature)
	if not geFeature then
		return 
	end
	
	local strJson = textFeature:command("get_draw_options","")
	if not strJson then
		return
	end		
	
	tbl = json.decode(strJson)
	if not tbl or not tbl["sdfTextureWidth"] or not tbl["sdfTextureHeight"] then
		return
	end
	
	local size = EffectSdk.vectorf()
	size:push_back(tbl["sdfTextureWidth"])
	size:push_back(tbl["sdfTextureHeight"])
	geFeature:setUniformVec2(text_pass_name,4,"uTexSize",size)	
	
end

-- got vertex uv and index data from textFeature. then we set them to renderFeature to render
local function setRenderOptions (strJson, feature)    
	if not strJson or not feature then        
		return 
	end		
    
    tbl = json.decode(strJson)
    local cmdTbl ={}
    cmdTbl[shadow_pass_name] = tbl
    cmdTbl[text_pass_name] = tbl
    local cmdStr = json.encode(cmdTbl)
    feature:command("set_properties",cmdStr)
    
    matTbl = tbl["MVPMatrix"]
--    printTable(matTbl)
    local geFeature = EffectSdk.castGeneralEffectFeature(feature)
 	if  matTbl and geFeature then
 		local mat = EffectSdk.vectorf()
 		for i=1, 16 do
 			mat:push_back(matTbl[i])
 		end
 		geFeature:setUniformMat4(shadow_pass_name,8,"MVPMatrix",mat)
 		geFeature:setUniformMat4(text_pass_name,8,"MVPMatrix",mat)
 	end    

    reforce_set_render_options = false
end

local function setRenderParameters(tbl, feature)
    if not tbl or not feature then    
		return 
	end	

    local geFeature = EffectSdk.castGeneralEffectFeature(feature)
    if not geFeature then
        return 
    end    
    
    if tbl["outlineWidth"] then
        local v = tbl["outlineWidth"]
        v = clamp(v,0.0, 1.0)
        geFeature:setUniformFloat(text_pass_name,3,"uOutlineWidth",v)
    end

    if tbl["outlineAlpha"] then
        local v = tbl["outlineAlpha"]
        v = clamp(v,0.0, 1.0)
        geFeature:setUniformFloat(text_pass_name,3,"uOutlineAlpha",v)
    end

    if tbl["shadowColor"] then
        local color = EffectSdk.vectorf()
        local v = tbl["shadowColor"]
        for i=1,4 do
            color:push_back(v[i])
        end
        geFeature:setUniformVec4(shadow_pass_name,6,"uShadowColor",color)
    end

    if tbl["shadowAlpha"] then
        local v = tbl["shadowAlpha"]
        v = clamp(v,0.0, 1.0)
        geFeature:setUniformFloat(shadow_pass_name,3,"uShadowAlpha",v)
    end

    if tbl["shadowOffset"] then        
        local v = tbl["shadowOffset"]
        local offset = EffectSdk.vectorf()
        for i=1,2 do
            offset:push_back(v[i])
        end
        geFeature:setUniformVec2(shadow_pass_name,4,"uShadowOffset",offset)
    end

    if tbl["shadowEnable"] then        
        local v = tbl["shadowEnable"]    
        local res = false
        if v == 1 then
            res = true
        end

        geFeature:setElementEnable(shadow_pass_name,res)
    end
    
    render_param_json = json.encode(tbl)
end

local function updateAnimationFeature(this, textFeature)
    if cur_animation_path and ani_load_status == 1 then    
        if last_animation_path then
            this:unloadFeature("Animation")
        end        
        this:loadFeature(cur_animation_path,"Animation","2DStickerV3",9000,true) 

        if exports.animation ~= nil then
            exports.animation.onstart(this)
        end
        
        ani_load_status = 0
    end   

    if ani_load_status == 2 then
        this:unloadFeature("Animation")
        ani_load_status = 0
        cur_animation_path = nil
        if exports.animation and textFeature then            
            exports.animation.onEnd(this, textFeature)
        end      
        exports.animation = nil
    end      
end

local function loadTextStyle(this)
    if text_state == 0 or cur_style_path == nil or style_load_status == 0 then
        return
    end

    local feature = this:loadFeature(cur_style_path,"FontRender/","GeneralEffect",4000,true)
    style_load_status = 0
    reforce_set_render_options = true
    if render_param_json ~= nil then
        local tbl = json.decode(render_param_json)
        setRenderParameters(tbl,feature)
    end   
end

local function updateRenderFeature(this, textFeature)
    local renderFeature = this:getFeature(ge_feature_name)
    if not renderFeature then
        return
    end

    if text_state == 0 then
        -- first frame textFeature not initialized. so we render next time
        text_state = 1        
        renderFeature:setFeatureStatus(EffectSdk.BEF_FEATURE_STATUS_ENABLED, false)  
        setSDFTextureSize(textFeature, renderFeature)        
    else   
    	
        local dirtyJson = textFeature:command("is_char_placements_dirty","")
        local tbl = json.decode(dirtyJson)

        if tbl["dirty"] or reforce_set_render_options then
            local strJson = textFeature:command("get_char_placements","")            
            if text_state == 1 then
                renderFeature:setFeatureStatus(EffectSdk.BEF_FEATURE_STATUS_ENABLED, true);   	
                text_state = 2;
            end                 	
            setRenderOptions(strJson, renderFeature)
        end
    end  
end

EventHandles = {
    handleEffectEvent = function (this, eventCode)
        if (init_state and eventCode == 1) then
            init_state = 0                          
        end
        return true
    end,

    
    handleUserEvent = function(this, json_string)
    	local textFeature = this:getFeature(text_feature_name)
    	if textFeature then
    		local user_data = json.decode(json_string)
    		if user_data["load_font"] then
        		local v = user_data["load_font"]
            	local json = string.format("{\"path\":\"%s\"}",v["path"])
            	textFeature:command("load_font",json)
            end
            
            -- implemented in OpenGL Thread
            if user_data["load_text_style"] then
        		local v = user_data["load_text_style"]
                local newPath = v["path"];
                if cur_style_path ~= newPath then
                    last_style_path = cur_style_path
                    cur_style_path = newPath
                    style_load_status = 1 
                end       
                
                if user_data["render_parameters"] then
                    render_param_json = json.encode(user_data["render_parameters"])
                else
                    render_param_json = nil;
                end
                
            end
            
            if user_data["update_text"] then
        		local v = user_data["update_text"]
            	local json = string.format("{\"text\":\"%s\"}",v["text"])
            	textFeature:command("update_text",json)
            end
            
            -- implemented in OpenGL Thread
            if user_data["load_animation"] then
        		local v = user_data["load_animation"]
        		local newPath = v["path"]
        		if cur_animation_path ~= newPath then
        			last_animation_path = cur_animation_path
        			cur_animation_path = newPath
        			ani_load_status = 1
        		end  
            end
            
            -- implemented in OpenGL Thread
            if user_data["unload_animation"] then
        		local v = user_data["unload_animation"]
        		ani_load_status = 2
            end
            
            
            if user_data["set_properties"] then
                textFeature:command("set_properties",json.encode(user_data["set_properties"]))
            end           
            
        end 
        
        local renderFeature = this:getFeature(ge_feature_name)
        if renderFeature then
            local user_data = json.decode(json_string)
            if user_data["render_parameters"] then
                setRenderParameters(user_data["render_parameters"],renderFeature)            
            end
        end

    end,

    -- handleUpdate will be called before we render one frame in OpenGL thread. or we call this: "PreRenderFrame"
    handleUpdate = function(this, milliSeconds)        
        if (init_state ~= 0) then 
            return true 
        end   
        
        local textFeature = this:getFeature(text_feature_name)
        
        updateAnimationFeature(this, textFeature)

        loadTextStyle(this)      

		
        if exports.animation then            
            exports.animation.ontimer(this, milliSeconds, textFeature)
        end      
        
        updateRenderFeature(this,textFeature)
              
    end,
    
    }
